# -*- coding: utf-8 -*-
#
# cwinpy documentation build configuration file (copied from that for bilby)
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

import cweqgen
from cweqgen.sphinx import (
    generate_equations_doc,
    generate_docstringexample_doc,
)

sys.path.insert(0, os.path.abspath("../cweqgen/"))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.intersphinx",
    "nbsphinx",
    "sphinxcontrib.napoleon",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
    "recommonmark",
    "sphinx_panels",
    "sphinx.ext.todo",
]

autosummary_generate = True
autodoc_docstring_signature = True
autodoc_default_options = {"members-order": "groupwise", "show-inheritance": True}
todo_include_todos = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = {
    ".rst": "restructuredtext",
}

# The master toctree document.
master_doc = "index"

# General information about the project.
project = u"cweqgen"
copyright = u"2022, Matthew Pitkin"
author = u"Matthew Pitkin"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
fullversion = cweqgen.__version__

# The short X.Y version.
version = ".".join(fullversion.split(".")[:2])

# The full version, including alpha/beta/rc tags.
release = fullversion

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "__pycache__",
]

add_function_parentheses = True
add_module_names = False
show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "cweqgendoc"

html_theme_options = {
    # "style_nav_header_background": "#e3e2e1",
    # "logo_only": True,
    "icon_links": [
        {
            "name": "GitLab",
            "url": "https://git.ligo.org/CW/software/CWEquationGenerator",
            "icon": "fab fa-gitlab",
        },
    ],
    "use_edit_page_button": True,
}

html_context = {
    "gitlab_url": "https://git.ligo.org",
    "gitlab_user": "CW/software",
    "gitlab_repo": "CWEquationGenerator",
    "gitlab_version": "main",
    "doc_path": "docs",
}

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "cweqgen.tex", u"cweqgen Documentation", u"Matthew Pitkin", "manual"),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "cweqgen", u"cweqgen Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "cweqgen",
        u"cweqgen Documentation",
        author,
        "cwinpy",
        "One line description of project.",
        "Miscellaneous",
    ),
]

numpydoc_show_class_members = False

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "astropy": ("https://docs.astropy.org/en/stable/", None),
    "sympy": ("https://docs.sympy.org/latest/", None),
    "matplotlib": ("https://matplotlib.org/", None),
}


# try and get dictionaries to display better https://stackoverflow.com/a/65195854/1862861
from pprint import pformat


def object_description(object) -> str:
    return pformat(object, indent=4, width=60)


from sphinx.util import inspect

inspect.object_description = object_description

generate_equations_doc("equations.rst")
generate_docstringexample_doc("docstringexample.txt")
